import re
from typing import List

from overrides import overrides

from allennlp.data.tokenizers.token_class import Token
from allennlp.data.tokenizers.tokenizer import Tokenizer


@Tokenizer.register("letters_digits")
class LettersDigitsTokenizer(Tokenizer):
    """
    A `Tokenizer` which keeps runs of (unicode) letters and runs of digits together, while
    every other non-whitespace character becomes a separate word.

    Registered as a `Tokenizer` with name "letters_digits".
    """

    @overrides
    def tokenize(self, text: str) -> List[Token]:
        # We use the [^\W\d_] pattern as a trick to match unicode letters
        tokens = [Token(m.group(), idx=m.start()) for m in re.finditer(r"[^\W\d_]+|\d+|\S", text)]
        return tokens



import os
from typing import Tuple, Union, Optional, TYPE_CHECKING, List, Any, Dict, Sequence

from fairscale.nn import FullyShardedDataParallel as FS_FSDP
from fairscale.nn.wrap import enable_wrap, wrap
from fairscale.nn.misc import FlattenParamsWrapper
from fairscale.optim.grad_scaler import GradScaler
from overrides import overrides
import torch
from torch.cuda import amp

from allennlp.nn.parallel.sharded_module_mixin import ShardedModuleMixin
from allennlp.nn.parallel.ddp_accelerator import (
    DdpAccelerator,
    DdpWrappedModel,
    StateDictType,
    LoadStateDictReturnType,
)

if TYPE_CHECKING:
    # To prevent circular imports
    from allennlp.models import Model


class _FSDP(FS_FSDP, ShardedModuleMixin):
    """
    Same as FairScale's [`FullyShardedDataParallel`]
    (https://fairscale.readthedocs.io/en/latest/api/nn/fsdp.html) but also implements
    the mixin methods from :class:`allennlp.nn.parallel.sharded_module_mixin.ShardedModuleMixin`.
    """

    @overrides
    def get_original_module(self) -> torch.nn.Module:
        module = self.module
        if isinstance(module, FlattenParamsWrapper):
            module = module.module
        return module


class FairScaleFsdpWrappedModel(DdpWrappedModel):
    """
    The wrapped model type returned from [`FairScaleFsdpWrappedModel.wrap_model`](#wrap_model).
    """

    @staticmethod
    @overrides
    def consolidate_sharded_state(
        sharded_state_files: Sequence[Union[str, os.PathLike]]
    ) -> StateDictType:
        shard_weights: List[StateDictType] = []
        shard_metadata: List[Dict[str, Any]] = []
        for path in sharded_state_files:
            shard_state = torch.load(path, map_location="cpu")
            shard_weights.append(shard_state["weights"])
            shard_metadata.append(shard_state["metadata"])
        return _FSDP.consolidate_shard_weights(shard_weights, shard_metadata)

    @overrides
    def load_state_dict(
        self, state_dict: StateDictType, strict: bool = True
    ) -> LoadStateDictReturnType:
        return self.model.load_local_state_dict(state_dict["weights"], strict=strict)  # type: ignore[operator]

    @overrides
    def state_dict(self, *args, **kwargs) -> StateDictType:
        weights = self.model.local_state_dict(*args, **kwargs)  # type: ignore[operator]
        metadata = self.model.local_metadata_dict()
        return {"weights": weights, "metadata": metadata}

    @overrides
    def clip_grad_norm_(self, max_norm: Union[float, int]) -> torch.Tensor:
        return self.model.clip_grad_norm_(max_norm)  # type: ignore[operator]

    @overrides
    def init_grad_scaler(self) -> amp.GradScaler:
        return GradScaler()


@DdpAccelerator.register("fairscale_fsdp")
class FairScaleFsdpAccelerator(DdpAccelerator):
    """
    A :class:`allennlp.nn.parallel.ddp_accelerator.DdpAccelerator` for FairScale's [`FullyShardedDataParallel`]
    (https://fairscale.readthedocs.io/en/latest/api/nn/fsdp.html).

    To save memory while initializing a model, you should call [`.wrap_module()`](#wrap_module) on submodules
    as they're created.

    See the :class:`allennlp.modules.transformer.t5.T5` class for an example of how to use this.
    """

    def __init__(
        self,
        *,
        mixed_precision: bool = False,
        reshard_after_forward: bool = True,
        flatten_parameters: bool = True,
        local_rank: Optional[int] = None,
        world_size: Optional[int] = None,
        cuda_device: Union[torch.device, int] = -1,
    ) -> None:
        super().__init__(local_rank=local_rank, world_size=world_size, cuda_device=cuda_device)
        self._fsdp_kwargs = {
            "compute_device": self.cuda_device,
            "mixed_precision": mixed_precision,
            "reshard_after_forward": reshard_after_forward,
            "flatten_parameters": flatten_parameters,
        }
        if mixed_precision:
            self._fsdp_kwargs["move_params_to_cpu"] = True
            self._fsdp_kwargs["clear_autocast_cache"] = True

    @overrides
    def wrap_model(self, model: "Model") -> Tuple["Model", DdpWrappedModel]:
        wrapped_model = _FSDP(
            model,
            **self._fsdp_kwargs,
        )
        if not self._fsdp_kwargs["mixed_precision"] and self.cuda_device != torch.device("cpu"):
            wrapped_model = wrapped_model.cuda()
        # `_FSDP._lazy_init()` may have been called already on submodules that were wrapped
        # (through `wrap_module()`), leading those submodules to think they are root submodules.
        # So we need to call `_FSDP._reset_lazy_init()` on any of these now.
        for module in wrapped_model.modules():
            if isinstance(module, _FSDP):
                module._reset_lazy_init()
        return model, FairScaleFsdpWrappedModel(
            wrapped_model,
            local_rank=self.local_rank,
            world_size=self.world_size,
        )

    @overrides
    def wrap_module(self, module: torch.nn.Module) -> torch.nn.Module:
        with enable_wrap(wrapper_cls=_FSDP, **self._fsdp_kwargs):
            wrapped_module = wrap(module)
        return wrapped_module



from typing import Any, Dict, Iterable, Iterator, Union, Optional
import itertools
import math

import torch
from overrides import overrides

from allennlp.common import util
from allennlp.data.batch import Batch
from allennlp.data.data_loaders.data_loader import DataLoader, TensorDict
from allennlp.data.data_loaders.multiprocess_data_loader import MultiProcessDataLoader
from allennlp.data.data_loaders.multitask_scheduler import MultiTaskScheduler
from allennlp.data.data_loaders.multitask_epoch_sampler import MultiTaskEpochSampler
from allennlp.data.dataset_readers.multitask import MultiTaskDatasetReader
from allennlp.data.instance import Instance
from allennlp.data.vocabulary import Vocabulary
import allennlp.nn.util as nn_util


def maybe_shuffle_instances(loader: DataLoader, shuffle: bool) -> Iterable[Instance]:
    if shuffle:
        return util.shuffle_iterable(loader.iter_instances())
    else:
        return loader.iter_instances()


@DataLoader.register("multitask")
class MultiTaskDataLoader(DataLoader):
    """
    A `DataLoader` intended for multi-task learning.  The basic idea is that you use a
    `MultiTaskDatasetReader`, which takes a dictionary of `DatasetReaders`, keyed by some name.  You
    use those same names for various parameters here, including the data paths that get passed to
    each reader.  We will load each dataset and iterate over instances in them using a
    `MultiTaskEpochSampler` and a `MultiTaskScheduler`.  The `EpochSampler` says how much to use
    from each dataset at each epoch, and the `Scheduler` orders the instances in the epoch however
    you want.  Both of these are designed to be used in conjunction with trainer `Callbacks`, if
    desired, to have the sampling and/or scheduling behavior be dependent on the current state of
    training.

    While it is not necessarily required, this `DatasetReader` was designed to be used alongside a
    `MultiTaskModel`, which can handle instances coming from different datasets.  If your datasets
    are similar enough (say, they are all reading comprehension datasets with the same format), or
    your model is flexible enough, then you could feasibly use this `DataLoader` with a normal,
    non-multitask `Model`.

    Registered as a `DataLoader` with name "multitask".

    # Parameters

    reader: `MultiTaskDatasetReader`
    data_path: `Dict[str, str]`
        One file per underlying dataset reader in the `MultiTaskDatasetReader`, which will be passed
        to those readers to construct one `DataLoader` per dataset.
    scheduler: `MultiTaskScheduler`, optional (default = `HomogeneousRoundRobinScheduler`)
        The `scheduler` determines how instances are ordered within an epoch.  By default, we'll
        select one batch of instances from each dataset in turn, trying to ensure as uniform a mix
        of datasets as possible.  Note that if your model can handle it, using a
        `RoundRobinScheduler` is likely better than a `HomogeneousRoundRobinScheduler` (because it
        does a better job mixing gradient signals from various datasets), so you may want to
        consider switching.  We use the homogeneous version as default because it should work for
        any allennlp model, while the non-homogeneous one might not.
    sampler: `MultiTaskEpochSampler`, optional (default = `None`)
        Only used if `instances_per_epoch` is not `None`. If we need to select a subset of the data
        for an epoch, this `sampler` will tell us with what proportion we should sample from each
        dataset.  For instance, we might want to focus more on datasets that are underperforming in
        some way, by having those datasets contribute more instances this epoch than other datasets.
    instances_per_epoch: `int`, optional (default = `None`)
        If not `None`, we will use this many instances per epoch of training, drawing from the
        underlying datasets according to the `sampler`.
    num_workers: `Dict[str, int]`, optional (default = `None`)
        Used when creating one `MultiProcessDataLoader` per dataset.  If you want non-default
        behavior for this parameter in the `DataLoader` for a particular dataset, pass the
        corresponding value here, keyed by the dataset name.
    max_instances_in_memory: `Dict[str, int]`, optional (default = `None`)
        Used when creating one `MultiProcessDataLoader` per dataset.  If you want non-default
        behavior for this parameter in the `DataLoader` for a particular dataset, pass the
        corresponding value here, keyed by the dataset name.
    start_method: `Dict[str, str]`, optional (default = `None`)
        Used when creating one `MultiProcessDataLoader` per dataset.  If you want non-default
        behavior for this parameter in the `DataLoader` for a particular dataset, pass the
        corresponding value here, keyed by the dataset name.
    instance_queue_size: `Dict[str, int]`, optional (default = `None`)
        Used when creating one `MultiProcessDataLoader` per dataset.  If you want non-default
        behavior for this parameter in the `DataLoader` for a particular dataset, pass the
        corresponding value here, keyed by the dataset name.
    instance_chunk_size: `Dict[str, int]`, optional (default = `None`)
        Used when creating one `MultiProcessDataLoader` per dataset.  If you want non-default
        behavior for this parameter in the `DataLoader` for a particular dataset, pass the
        corresponding value here, keyed by the dataset name.
    shuffle: `bool`, optional (default = `True`)
        If `False`, we will not shuffle the instances that come from each underlying data loader.
        You almost certainly never want to use this except when debugging.
    cuda_device: `Optional[Union[int, str, torch.device]]`, optional (default = `None`)
        If given, batches will automatically be put on this device.

        !!! Note
            This should typically not be set in an AllenNLP configuration file. The `Trainer`
            will automatically call [`set_target_device()`](#set_target_device) before iterating
            over batches.
    """

    def __init__(
        self,
        reader: MultiTaskDatasetReader,
        data_path: Dict[str, str],
        scheduler: MultiTaskScheduler,
        *,
        sampler: MultiTaskEpochSampler = None,
        instances_per_epoch: int = None,
        num_workers: Dict[str, int] = None,
        max_instances_in_memory: Dict[str, int] = None,
        start_method: Dict[str, str] = None,
        instance_queue_size: Dict[str, int] = None,
        instance_chunk_size: Dict[str, int] = None,
        shuffle: bool = True,
        cuda_device: Optional[Union[int, str, torch.device]] = None,
    ) -> None:
        self.readers = reader.readers
        self.data_paths = data_path
        self.scheduler = scheduler
        self.sampler = sampler
        self.cuda_device: Optional[torch.device] = None
        if cuda_device is not None:
            if not isinstance(cuda_device, torch.device):
                self.cuda_device = torch.device(cuda_device)
            else:
                self.cuda_device = cuda_device

        self._instances_per_epoch = instances_per_epoch
        self._shuffle = shuffle

        if instances_per_epoch is not None and sampler is None:
            raise ValueError(
                "You must provide an EpochSampler if you want to not use all instances every epoch."
            )

        self._num_workers = num_workers or {}
        self._max_instances_in_memory = max_instances_in_memory or {}
        self._start_method = start_method or {}
        self._instance_queue_size = instance_queue_size or {}
        self._instance_chunk_size = instance_chunk_size or {}

        if self.readers.keys() != self.data_paths.keys():
            raise ValueError(
                f"Mismatch between readers ({self.readers.keys()}) and data paths "
                f"({self.data_paths.keys()})"
            )
        self._loaders = {key: self._make_data_loader(key) for key in self.readers}

        # This stores our current iterator with each dataset, so we don't just iterate over the
        # first k instances every epoch if we're using instances_per_epoch.  We'll grab instances
        # from here each epoch, and refresh it when it runs out.  We only use this in the case that
        # instances_per_epoch is not None, but these iterators are lazy, so always creating them
        # doesn't hurt anything.
        self._iterators: Dict[str, Iterator[Instance]] = {
            # NOTE: The order in which we're calling these iterator functions is important.  We want
            # an infinite iterator over the data, but we want the order in which we iterate over the
            # data to be different at every epoch.  The cycle function will give us an infinite
            # iterator, and it will call the lambda function each time it runs out of instances,
            # which will produce a new shuffling of the dataset.
            key: util.cycle_iterator_function(
                # This default argument to the lambda function is necessary to create a new scope
                # for the loader variable, so a _different_ loader gets saved for every iterator.
                # Dictionary comprehensions don't create new scopes in python.  If you don't have
                # this loader, you end up with `loader` always referring to the last loader in the
                # iteration... mypy also doesn't know what to do with this, for some reason I can't
                # figure out.
                lambda l=loader: maybe_shuffle_instances(l, self._shuffle)  # type: ignore
            )
            for key, loader in self._loaders.items()
        }

    @overrides
    def __len__(self) -> int:
        if self._instances_per_epoch is None:
            # This will raise a TypeError if any of the underlying loaders doesn't have a length,
            # which is actually what we want.
            return self.scheduler.count_batches(
                {dataset: len(loader) for dataset, loader in self._loaders.items()}
            )
        else:
            return self.scheduler.count_batches(
                {dataset: self._instances_per_epoch for dataset in self._loaders.keys()}
            )

    @overrides
    def __iter__(self) -> Iterator[TensorDict]:
        epoch_instances = self._get_instances_for_epoch()
        return (
            nn_util.move_to_device(
                Batch(instances).as_tensor_dict(),
                -1 if self.cuda_device is None else self.cuda_device,
            )
            for instances in self.scheduler.batch_instances(epoch_instances)
        )

    @overrides
    def iter_instances(self) -> Iterator[Instance]:
        # The only external contract for this method is that it iterates over instances
        # individually; it doesn't actually specify anything about batching or anything else.  The
        # implication is that you iterate over all instances in the dataset, in an arbitrary order.
        # The only external uses of this method are in vocabulary construction (the
        # MultiProcessDataLoader uses this function internally when constructing batches, but that's
        # an implementation detail).
        #
        # So, the only thing we need to do here is iterate over all instances from all datasets, and
        # that's sufficient.  We won't be using this for batching, because that requires some
        # complex, configurable scheduling.
        #
        # The underlying data loaders here could be using multiprocessing; we don't need to worry
        # about that in this class. Caching is also handled by the underlying data loaders.
        for loader in self._loaders.values():
            yield from loader.iter_instances()

    @overrides
    def index_with(self, vocab: Vocabulary) -> None:
        for loader in self._loaders.values():
            loader.index_with(vocab)

    @overrides
    def set_target_device(self, device: torch.device) -> None:
        self.cuda_device = device

    def _get_instances_for_epoch(self) -> Dict[str, Iterable[Instance]]:
        if self._instances_per_epoch is None:
            return {
                key: maybe_shuffle_instances(loader, self._shuffle)
                for key, loader in self._loaders.items()
            }
        if self.sampler is None:
            # We already checked for this in the constructor, so this should never happen unless you
            # modified the object after creation. But mypy is complaining, so here's another check.
            raise ValueError(
                "You must specify an EpochSampler if self._instances_per_epoch is not None."
            )
        dataset_proportions = self.sampler.get_task_proportions(self._loaders)
        proportion_sum = sum(dataset_proportions.values())
        num_instances_per_dataset = {
            key: math.floor(proportion * self._instances_per_epoch / proportion_sum)
            for key, proportion in dataset_proportions.items()
        }
        return {
            key: itertools.islice(self._iterators[key], num_instances)
            for key, num_instances in num_instances_per_dataset.items()
        }

    def _make_data_loader(self, key: str) -> MultiProcessDataLoader:
        kwargs: Dict[str, Any] = {
            "reader": self.readers[key],
            "data_path": self.data_paths[key],
            # We don't load batches from this data loader, only instances, but we have to set
            # something for the batch size, so we set 1.
            "batch_size": 1,
        }
        if key in self._num_workers:
            kwargs["num_workers"] = self._num_workers[key]
        if key in self._max_instances_in_memory:
            kwargs["max_instances_in_memory"] = self._max_instances_in_memory[key]
        if key in self._start_method:
            kwargs["start_method"] = self._start_method[key]
        return MultiProcessDataLoader(**kwargs)



"""
Subcommand for building a vocabulary from a training config.
"""

import argparse
import json
import logging
import os
import tarfile
import tempfile

from overrides import overrides

from allennlp.commands.subcommand import Subcommand
from allennlp.common.file_utils import CacheFile
from allennlp.common.params import Params
from allennlp.training.util import make_vocab_from_params


logger = logging.getLogger(__name__)


@Subcommand.register("build-vocab")
class BuildVocab(Subcommand):
    @overrides
    def add_subparser(self, parser: argparse._SubParsersAction) -> argparse.ArgumentParser:
        description = """Build a vocabulary from an experiment config file."""
        subparser = parser.add_parser(self.name, description=description, help=description)

        subparser.add_argument("param_path", type=str, help="path to an experiment config file")

        subparser.add_argument(
            "output_path", type=str, help="path to save the vocab tar.gz file to"
        )

        subparser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="force write if the output_path already exists",
        )

        subparser.add_argument(
            "-o",
            "--overrides",
            type=str,
            default="",
            help=(
                "a json(net) structure used to override the experiment configuration, e.g., "
                "'{\"vocabulary.min_count.labels\": 10}'.  Nested parameters can be specified either"
                " with nested dictionaries or with dot syntax."
            ),
        )

        subparser.set_defaults(func=build_vocab_from_args)

        return subparser


def build_vocab_from_args(args: argparse.Namespace):
    if not args.output_path.endswith(".tar.gz"):
        raise ValueError("param 'output_path' should end with '.tar.gz'")

    if os.path.exists(args.output_path) and not args.force:
        raise RuntimeError(f"{args.output_path} already exists. Use --force to overwrite.")

    output_directory = os.path.dirname(args.output_path)
    if len(output_directory) > 0:
        os.makedirs(output_directory, exist_ok=True)

    params = Params.from_file(args.param_path)

    with tempfile.TemporaryDirectory() as temp_dir:
        # Serializes the vocab to 'tempdir/vocabulary'.
        make_vocab_from_params(params, temp_dir)

        # The CacheFile context manager gives us a temporary file to write to.
        # On a successful exit from the context, it will rename the temp file to
        # the target `output_path`.
        with CacheFile(args.output_path, suffix=".tar.gz") as temp_archive:
            logger.info("Archiving vocabulary to %s", args.output_path)

            with tarfile.open(temp_archive.name, "w:gz") as archive:
                vocab_dir = os.path.join(temp_dir, "vocabulary")
                for fname in os.listdir(vocab_dir):
                    if fname.endswith(".lock"):
                        continue
                    archive.add(os.path.join(vocab_dir, fname), arcname=fname)

    print(f"Success! Vocab saved to {args.output_path}")
    print('You can now set the "vocabulary" entry of your training config to:')
    print(json.dumps({"type": "from_files", "directory": os.path.abspath(args.output_path)}))



"""
*AllenNLP Tango is an experimental API and parts of it might change or disappear
every time we release a new version.*
"""

import datasets
from allennlp.tango.dataset import DatasetDict
from allennlp.tango.step import Step


@Step.register("hf_dataset")
class HuggingfaceDataset(Step):
    """This steps reads a huggingface dataset and returns it in `AllenNlpDataset` format."""

    DETERMINISTIC = True
    VERSION = "001"
    CACHEABLE = False  # These are already cached by huggingface.

    def run(self, dataset_name: str) -> DatasetDict:  # type: ignore
        """Reads and returns a huggingface dataset. `dataset_name` is the name of the dataset."""
        return DatasetDict(datasets.load_dataset(dataset_name), None, {"source": "huggingface"})



from copy import deepcopy
import torch


def replicate_layers(layer: torch.nn.Module, num_copies: int):
    """
    # Parameters
            layer (torch.nn.Module) - The torch layer that needs to be replicated.
            num_copies (int) - Number of copies to create.

    # Returns
            A ModuleList that contains `num_copies` of the `layer`.
    """
    return torch.nn.ModuleList([deepcopy(layer) for _ in range(num_copies)])



"""
This is a docstring.

And this is a multi-line line: [http://example.com]
(https://example.com/blah/blah/blah.html).
"""

from dataclasses import dataclass

SOME_GLOBAL_VAR = "Ahhhh I'm a global var!!"
"""
This is a global var.
"""


def func_with_no_args():
    """
    This function has no args.
    """
    return None


def func_with_args(a: int, b: int, c: int = 3) -> int:
    """
    This function has some args.

    # Parameters

    a : `int`
        A number.
    b : `int`
        Another number.
    c : `int`, optional (default = `3`)
        Yet another number.

    Notes
    -----

    These are some notes.

    # Returns

    `int`
        The result of `a + b * c`.
    """
    return a + b * c


class SomeClass:
    """
    I'm a class!

    # Parameters

    x : `float`
        This attribute is called `x`.
    """

    some_class_level_variable = 1
    """
    This is how you document a class-level variable.
    """

    some_class_level_var_with_type: int = 1

    def __init__(self) -> None:
        self.x = 1.0

    def _private_method(self) -> None:
        """
        Private methods should not be included in documentation.
        """
        pass

    def some_method(self) -> None:
        """
        I'm a method!

        But I don't do anything.

        # Returns

        `None`
        """
        return None

    def method_with_alternative_return_section(self) -> int:
        """
        Another method.

        # Returns

        A completely arbitrary number.
        """
        return 3

    def method_with_alternative_return_section3(self) -> int:
        """
        Another method.

        # Returns

        number : `int`
            A completely arbitrary number.
        """
        return 3


class AnotherClassWithReallyLongConstructor:
    def __init__(
        self,
        a_really_long_argument_name: int = 0,
        another_long_name: float = 2,
        these_variable_names_are_terrible: str = "yea I know",
        **kwargs,
    ) -> None:
        self.a = a_really_long_argument_name
        self.b = another_long_name
        self.c = these_variable_names_are_terrible
        self.other = kwargs


@dataclass
class ClassWithDecorator:
    x: int


class _PrivateClass:
    def public_method_on_private_class(self):
        """
        This should not be documented since the class is private.
        """
        pass



from allennlp.nn.parallel.sharded_module_mixin import ShardedModuleMixin
from allennlp.nn.parallel.ddp_accelerator import (
    DdpAccelerator,
    DdpWrappedModel,
    TorchDdpAccelerator,
)
from allennlp.nn.parallel.fairscale_fsdp_accelerator import (
    FairScaleFsdpAccelerator,
    FairScaleFsdpWrappedModel,
)



"""
Subcommand for running Tango experiments
"""

import argparse
import logging
import os
from os import PathLike
from pathlib import Path
from typing import Union, Dict, Any, List, Optional
import warnings

from overrides import overrides

from allennlp.commands.subcommand import Subcommand
from allennlp.common.params import Params
from allennlp.common import logging as common_logging
from allennlp.common import util as common_util
from allennlp.tango.step import step_graph_from_params, tango_dry_run
from allennlp.tango.step import DirectoryStepCache

logger = logging.getLogger(__name__)


@Subcommand.register("tango")
class Tango(Subcommand):
    @overrides
    def add_subparser(self, parser: argparse._SubParsersAction) -> argparse.ArgumentParser:
        description = """Run a tango experiment file."""
        subparser = parser.add_parser(self.name, description=description, help=description)

        subparser.add_argument("config_path", type=str, help="path to a Tango experiment file")

        subparser.add_argument(
            "-s",
            "--serialization-dir",
            required=True,
            type=str,
            help="directory in which to save the results of the steps",
        )

        subparser.add_argument(
            "-o",
            "--overrides",
            type=str,
            default="",
            help=(
                "a json(net) structure used to override the experiment configuration, e.g., "
                "'{\"vocabulary.min_count.labels\": 10}'. Nested parameters can be specified either"
                " with nested dictionaries or with dot syntax."
            ),
        )

        subparser.add_argument(
            "--dry-run",
            action="store_true",
            help="Only show what would run. Don't run anything.",
        )

        subparser.add_argument(
            "--file-friendly-logging",
            action="store_true",
            default=False,
            help="outputs tqdm status on separate lines and slows tqdm refresh rate",
        )

        subparser.set_defaults(func=run_tango_from_args)

        return subparser


def run_tango_from_args(args: argparse.Namespace):
    warnings.warn(
        "AllenNLP Tango is an experimental API and parts of it might change or disappear "
        "every time we release a new version."
    )
    run_tango_from_file(
        tango_filename=args.config_path,
        serialization_dir=args.serialization_dir,
        overrides=args.overrides,
        include_package=args.include_package,
        dry_run=args.dry_run,
        file_friendly_logging=args.file_friendly_logging,
    )


def run_tango_from_file(
    tango_filename: Union[str, PathLike],
    serialization_dir: Union[str, PathLike],
    overrides: Union[str, Dict[str, Any]] = "",
    include_package: Optional[List[str]] = None,
    dry_run: bool = False,
    file_friendly_logging: bool = False,
):
    params = Params.from_file(tango_filename, overrides)
    return run_tango(
        params=params,
        serialization_dir=serialization_dir,
        include_package=include_package,
        dry_run=dry_run,
        file_friendly_logging=file_friendly_logging,
    )


def run_tango(
    params: Params,
    serialization_dir: Union[str, PathLike],
    include_package: Optional[List[str]] = None,
    dry_run: bool = False,
    file_friendly_logging: bool = False,
):
    common_logging.FILE_FRIENDLY_LOGGING = file_friendly_logging

    if include_package is not None:
        for package_name in include_package:
            common_util.import_module_and_submodules(package_name)

    common_util.prepare_environment(params)

    step_graph = step_graph_from_params(params.pop("steps"))

    serialization_dir = Path(serialization_dir)
    serialization_dir.mkdir(parents=True, exist_ok=True)
    step_cache = DirectoryStepCache(serialization_dir / "step_cache")

    if dry_run:
        for step, cached in tango_dry_run(
            (s for s in step_graph.values() if not s.only_if_needed), step_cache
        ):
            if cached:
                print(f"Getting {step.name} from cache")
            else:
                print(f"Computing {step.name}")
    else:
        # remove symlinks to old results
        for filename in serialization_dir.glob("*"):
            if filename.is_symlink():
                relative_target = os.readlink(filename)
                if not relative_target.startswith("step_cache/"):
                    continue
                logger.info(
                    f"Removing symlink '{filename.name}' to previous result {relative_target}"
                )
                filename.unlink()

        # produce results
        for name, step in step_graph.items():
            if not step.only_if_needed:
                step.ensure_result(step_cache)

        # symlink everything that has been computed
        for name, step in step_graph.items():
            if step in step_cache:
                step_link = serialization_dir / name
                step_link.unlink(missing_ok=True)
                step_link.symlink_to(
                    step_cache.path_for_step(step).relative_to(serialization_dir),
                    target_is_directory=True,
                )
                print(f'The output for "{name}" is in {step_link}.')



import os

from allennlp.common.file_utils import CACHE_DIRECTORY
from allennlp.common.file_utils import filename_to_url


def main():
    print(f"Looking for datasets in {CACHE_DIRECTORY}...")
    if not os.path.exists(CACHE_DIRECTORY):
        print("Directory does not exist.")
        print("No cached datasets found.")

    cached_files = os.listdir(CACHE_DIRECTORY)

    if not cached_files:
        print("Directory is empty.")
        print("No cached datasets found.")

    for filename in cached_files:
        if not filename.endswith("json"):
            url, etag = filename_to_url(filename)
            print("Filename: %s" % filename)
            print("Url: %s" % url)
            print("ETag: %s" % etag)
            print()


if __name__ == "__main__":
    main()



from overrides import overrides

import torch

from allennlp.modules.seq2vec_encoders.seq2vec_encoder import Seq2VecEncoder
from allennlp.nn.util import get_lengths_from_binary_sequence_mask


@Seq2VecEncoder.register("boe")
@Seq2VecEncoder.register("bag_of_embeddings")
class BagOfEmbeddingsEncoder(Seq2VecEncoder):
    """
    A `BagOfEmbeddingsEncoder` is a simple [`Seq2VecEncoder`](./seq2vec_encoder.md) which simply sums
    the embeddings of a sequence across the time dimension. The input to this module is of shape
    `(batch_size, num_tokens, embedding_dim)`, and the output is of shape `(batch_size, embedding_dim)`.

    Registered as a `Seq2VecEncoder` with name "bag_of_embeddings" and "boe".

    # Parameters

    embedding_dim : `int`, required
        This is the input dimension to the encoder.
    averaged : `bool`, optional (default=`False`)
        If `True`, this module will average the embeddings across time, rather than simply summing
        (ie. we will divide the summed embeddings by the length of the sentence).
    """

    def __init__(self, embedding_dim: int, averaged: bool = False) -> None:
        super().__init__()
        self._embedding_dim = embedding_dim
        self._averaged = averaged

    @overrides
    def get_input_dim(self) -> int:
        return self._embedding_dim

    @overrides
    def get_output_dim(self) -> int:
        return self._embedding_dim

    def forward(self, tokens: torch.Tensor, mask: torch.BoolTensor = None):
        if mask is not None:
            tokens = tokens * mask.unsqueeze(-1)

        # Our input has shape `(batch_size, num_tokens, embedding_dim)`, so we sum out the `num_tokens`
        # dimension.
        summed = tokens.sum(1)

        if self._averaged:
            if mask is not None:
                lengths = get_lengths_from_binary_sequence_mask(mask)
                length_mask = lengths > 0

                # Set any length 0 to 1, to avoid dividing by zero.
                lengths = torch.max(lengths, lengths.new_ones(1))
            else:
                lengths = tokens.new_full((1,), fill_value=tokens.size(1))
                length_mask = None

            summed = summed / lengths.unsqueeze(-1).float()

            if length_mask is not None:
                summed = summed * (length_mask > 0).unsqueeze(-1)

        return summed



import torch
from allennlp.data.vocabulary import Vocabulary
from allennlp.models.model import Model


class FakeModelForTestingNormalizationBiasVerification(Model):
    def __init__(self, use_bias=True):
        super().__init__(vocab=Vocabulary())
        self.conv = torch.nn.Conv2d(3, 5, kernel_size=1, bias=use_bias)
        self.bn = torch.nn.BatchNorm2d(5)

    def forward(self, x):
        # x: (B, 3, H, W)
        out = self.bn(self.conv(x))
        return {"loss": out.sum()}



"""
Wrappers for PyTorch's Learning Rate Schedulers so that they work with AllenNLP
"""
from typing import List, Union

import torch
from allennlp.training.optimizers import Optimizer
from allennlp.training.learning_rate_schedulers.learning_rate_scheduler import (
    LearningRateScheduler,
    _PyTorchLearningRateSchedulerWrapper,
    _PyTorchLearningRateSchedulerWithMetricsWrapper,
)


@LearningRateScheduler.register("step")
class StepLearningRateScheduler(_PyTorchLearningRateSchedulerWrapper):
    """
    Wrapper for [PyTorch's `StepLR`](
    https://pytorch.org/docs/stable/generated/
    torch.optim.lr_scheduler.StepLR.html#torch.optim.lr_scheduler.StepLR
    ).

    Registered as a `LearningRateScheduler` with name "step".  The "optimizer"
    argument does not get an entry in a configuration file for the object.

    # Parameters
    optimizer : `torch.optim.Optimizer`
        This argument does not get an entry in a configuration file for the
        object.
    step_size : `int`
        Period of learning rate decay.
    gamma: `float`, optional (default = `0.1`)
        Multiplicative factor of learning rate decay.
    last_epoch : `int`, optional (default=`-1`)
        The index of the last epoch. This is used when restarting.


    # Example

    Config for using the `StepLearningRateScheduler` Learning Rate Scheduler
    with `step_size` of `100` and `gamma` set `0.2`.

    ```json
    {
        ...
       "trainer":{
            ...
            "learning_rate_scheduler": {
                "type": "step",
                "step_size": 100,
                "gamma": 0.2
            },
            ...
       }
    }
    ```
    Note that you do NOT pass a `optimizer` key to the Learning rate scheduler.

    """

    def __init__(
        self, optimizer: Optimizer, step_size: int, gamma: float = 0.1, last_epoch: int = -1
    ) -> None:
        """

        Args:

        """

        lr_scheduler = torch.optim.lr_scheduler.StepLR(
            optimizer=optimizer, step_size=step_size, gamma=gamma, last_epoch=last_epoch
        )
        super().__init__(lr_scheduler)


@LearningRateScheduler.register("multi_step")
class MultiStepLearningRateScheduler(_PyTorchLearningRateSchedulerWrapper):
    """
    Wrapper for [PyTorch's `MultiStepLR`](
    https://pytorch.org/docs/stable/
    generated/torch.optim.lr_scheduler.MultiStepLR.html#torch.optim.lr_scheduler.MultiStepLR
    ).

    Registered as a `LearningRateScheduler` with name "multi_step".  The "optimizer" argument does
    not get an entry in a configuration file for the object.

    # Example

    Config for using the `MultiStepLearningRateScheduler` Learning Rate
    Scheduler with `milestones` of `[10,20,40]` and `gamma` set `0.2`.

    ```json
    {
        ...
       "trainer":{
            ...
            "learning_rate_scheduler": {
                "type": "multi_step",
                "milestones": [10,20,40],
                "gamma": 0.2
            },
            ...
       }
    }
    ```
    Note that you do NOT pass a `optimizer` key to the Learning rate scheduler.
    """

    def __init__(
        self, optimizer: Optimizer, milestones: List[int], gamma: float = 0.1, last_epoch: int = -1
    ) -> None:
        lr_scheduler = torch.optim.lr_scheduler.MultiStepLR(
            optimizer=optimizer, milestones=milestones, gamma=gamma, last_epoch=last_epoch
        )
        super().__init__(lr_scheduler)


@LearningRateScheduler.register("exponential")
class ExponentialLearningRateScheduler(_PyTorchLearningRateSchedulerWrapper):
    """
    Wrapper for [PyTorch's `ExponentialLR`](
    https://pytorch.org/docs/stable/generated/
    torch.optim.lr_scheduler.ExponentialLR.html#torch.optim.lr_scheduler.ExponentialLR).

    Registered as a `LearningRateScheduler` with name "exponential".  The "optimizer" argument does
    not get an entry in a configuration file for the object.

    # Example

    Config for using the `ExponentialLearningRateScheduler` Learning Rate
    Scheduler with `gamma` set `0.2`.

    ```json
    {
        ...
       "trainer":{
            ...
            "learning_rate_scheduler": {
                "type": "exponential",
                "gamma": 0.2
            },
            ...
       }
    }
    ```
    Note that you do NOT pass a `optimizer` key to the Learning rate scheduler.
    """

    def __init__(self, optimizer: Optimizer, gamma: float = 0.1, last_epoch: int = -1) -> None:
        lr_scheduler = torch.optim.lr_scheduler.ExponentialLR(
            optimizer=optimizer, gamma=gamma, last_epoch=last_epoch
        )
        super().__init__(lr_scheduler)


@LearningRateScheduler.register("reduce_on_plateau")
class ReduceOnPlateauLearningRateScheduler(_PyTorchLearningRateSchedulerWithMetricsWrapper):
    """
    Wrapper for [PyTorch's `ReduceLROnPlateau`](https://pytorch.org/docs/stable/generated/
    torch.optim.lr_scheduler.ReduceLROnPlateau.html#torch.optim.lr_scheduler.ReduceLROnPlateau).

    Registered as a `LearningRateScheduler` with name "reduce_on_plateau".  The
    "optimizer" argument does not get an entry in a configuration file for the
    object.

     # Example

    Config for using the `ReduceOnPlateauLearningRateScheduler` Learning Rate
    Scheduler with the following `init` arguments:

    * `mode="max"`
    * `factor=0.2`
    * `patience=5`
    * `threshold=5e-3`
    * `threshold_mode="abs"`
    * `cooldown=2`
    * `min_lr=1e-12`
    * `eps=1e-10`

    ```json
    {
        ...
       "trainer":{
            ...
            "learning_rate_scheduler": {
                "type": "reduce_on_plateau",
                "mode": "max",
                "factor": 0.2,
                "patience": 5,
                "threshold": 5e-3,
                "threshold_mode": "abs",
                "cooldown": 2,
                "min_lr": 1e-12,
                "eps": 1e-10
            },
            ...
       }
    }
    ```
    Note that you do NOT pass a `optimizer` key to the Learning rate scheduler.
    """

    def __init__(
        self,
        optimizer: Optimizer,
        mode: str = "min",
        factor: float = 0.1,
        patience: int = 10,
        verbose: bool = False,
        threshold_mode: str = "rel",
        threshold: float = 1e-4,
        cooldown: int = 0,
        min_lr: Union[float, List[float]] = 0,
        eps: float = 1e-8,
    ) -> None:
        lr_scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            optimizer=optimizer,
            mode=mode,
            factor=factor,
            patience=patience,
            verbose=verbose,
            threshold_mode=threshold_mode,
            threshold=threshold,
            cooldown=cooldown,
            min_lr=min_lr,
            eps=eps,
        )
        super().__init__(lr_scheduler)



import pytest
import sqlite3
from unittest.mock import call, Mock

from allennlp.common.testing import AllenNlpTestCase

from scripts.ai2_internal.resume_daemon import (
    BeakerStatus,
    create_table,
    handler,
    logger,
    resume,
    start_autoresume,
)

# Don't spam the log in tests.
logger.removeHandler(handler)


class ResumeDaemonTest(AllenNlpTestCase):
    def setup_method(self):
        super().setup_method()
        self.connection = sqlite3.connect(":memory:")
        create_table(self.connection)

    def test_create_beaker_status_works(self):
        status = BeakerStatus("stopped")
        assert status.name == "stopped"

    def test_create_beaker_status_throws(self):
        with pytest.raises(ValueError):
            status = BeakerStatus("garbage")
            assert status.name == "garbage"

    def test_does_nothing_on_empty_db(self):
        beaker = Mock()
        resume(self.connection, beaker)
        assert not beaker.method_calls

    def test_does_not_resume_a_running_experiment(self):
        beaker = Mock()
        experiment_id = "foo"
        start_autoresume(self.connection, experiment_id, 5)
        beaker.get_status.return_value = BeakerStatus.running
        resume(self.connection, beaker)
        beaker.get_status.assert_called()
        assert len(beaker.method_calls) == 1

    def test_does_not_resume_a_finished_experiment(self):
        beaker = Mock()
        experiment_id = "foo"
        start_autoresume(self.connection, experiment_id, 5)
        beaker.get_status.return_value = BeakerStatus.succeeded
        resume(self.connection, beaker)
        beaker.get_status.assert_called()
        assert len(beaker.method_calls) == 1

    def test_does_resume_a_preempted_experiment(self):
        beaker = Mock()
        experiment_id = "foo"
        start_autoresume(self.connection, experiment_id, 5)
        beaker.get_status.return_value = BeakerStatus.preempted
        beaker.resume.return_value = "foo2"
        resume(self.connection, beaker)
        beaker.get_status.assert_called()
        beaker.resume.assert_called()
        assert len(beaker.method_calls) == 2

    def test_respects_upper_bound_on_resumes(self):
        beaker = Mock()
        experiment_id = "foo"
        start_autoresume(self.connection, experiment_id, 5)
        beaker.get_status.return_value = BeakerStatus.preempted
        for i in range(10):
            beaker.resume.return_value = f"foo{i}"
            resume(self.connection, beaker)
        calls = [
            call.get_status("foo"),
            call.resume("foo"),
            call.get_status("foo0"),
            call.resume("foo0"),
            call.get_status("foo1"),
            call.resume("foo1"),
            call.get_status("foo2"),
            call.resume("foo2"),
            call.get_status("foo3"),
            call.resume("foo3"),
            call.get_status("foo4"),
        ]
        beaker.assert_has_calls(calls)

    def test_handles_a_realistic_scenario(self):
        beaker = Mock()
        experiment_id = "foo"
        start_autoresume(self.connection, experiment_id, 5)
        beaker.get_status.return_value = BeakerStatus.preempted
        for i in range(10):
            beaker.resume.return_value = f"foo{i}"
            if i == 2:
                beaker.get_status.return_value = BeakerStatus.succeeded
            resume(self.connection, beaker)
        calls = [
            call.get_status("foo"),
            call.resume("foo"),
            call.get_status("foo0"),
            call.resume("foo0"),
            call.get_status("foo1"),
        ]
        beaker.assert_has_calls(calls)



from collections import OrderedDict
from typing import Tuple

from torch import nn, FloatTensor, IntTensor
import torchvision

from allennlp.common.registrable import Registrable


class GridEmbedder(nn.Module, Registrable):
    """
    A `GridEmbedder` takes a batch of images as a tensor with shape
    `(batch_size, color_channels, height, width)`, and returns an ordered dictionary
    of tensors with shape `(batch_size, *)`, each representing a specific feature.
    """

    def forward(self, images: FloatTensor, sizes: IntTensor) -> "OrderedDict[str, FloatTensor]":
        raise NotImplementedError()

    def get_feature_names(self) -> Tuple[str, ...]:
        """
        Returns the feature names, in order, i.e. the keys of the ordered output
        dictionary from `.forward()`.
        """
        raise NotImplementedError()


@GridEmbedder.register("null")
class NullGridEmbedder(GridEmbedder):
    """
    A `GridEmbedder` that returns the input image as given.
    """

    def forward(self, images: FloatTensor, sizes: IntTensor) -> "OrderedDict[str, FloatTensor]":
        out = OrderedDict()
        out["0"] = images
        return out

    def get_feature_names(self) -> Tuple[str, ...]:
        return ("0",)


@GridEmbedder.register("resnet_backbone")
class ResnetBackbone(GridEmbedder):
    """
    Runs an image through [ResNet](https://api.semanticscholar.org/CorpusID:206594692),
    as implemented by [torchvision](https://pytorch.org/docs/stable/torchvision/models.html).
    """

    def __init__(self) -> None:
        super().__init__()
        detection_model = torchvision.models.detection.fasterrcnn_resnet50_fpn(pretrained=True)
        self.backbone = detection_model.backbone
        # Don't need the rest of this.
        del detection_model
        self.feature_names = tuple(
            [
                self.backbone.body.return_layers[key]
                for key in self.backbone.body.keys()
                if key in self.backbone.body.return_layers
            ]
            + ["pool"]
        )

    def forward(self, images: FloatTensor, sizes: IntTensor) -> "OrderedDict[str, FloatTensor]":
        return self.backbone(images)

    def get_feature_names(self) -> Tuple[str, ...]:
        return self.feature_names



from typing import Union, TYPE_CHECKING

import torch

from allennlp.common import FromParams
from allennlp.modules.transformer.activation_layer import ActivationLayer

if TYPE_CHECKING:
    from transformers.configuration_utils import PretrainedConfig


class TransformerPooler(ActivationLayer, FromParams):

    _pretrained_relevant_module = ["pooler", "bert.pooler", "roberta.pooler"]

    def __init__(
        self,
        hidden_size: int,
        intermediate_size: int,
        activation: Union[str, torch.nn.Module] = "relu",
    ):
        super().__init__(hidden_size, intermediate_size, activation, pool=True)

    @classmethod
    def _from_config(cls, config: "PretrainedConfig", **kwargs):
        return cls(config.hidden_size, config.hidden_size, "tanh")  # BERT has this hardcoded



from allennlp.models.model import Model


class Head(Model):
    """
    A `Head` is a `Model` that takes _already encoded input_ and typically does simple computation
    before returning a loss.

    There isn't currently any difference in API between a `Model` and a `Head`, but we have this
    separate type as both a signaling mechanism for what to expect when looking at a `Head` class,
    and so that we can use this as a more informative type annotation when building models that use
    `Heads` as inputs.

    One additional consideration in a `Head` is that `make_output_human_readable` needs to account
    for the case where it gets called without first having `forward` be called on the head.  This is
    because at the point where we call `make_output_human_readable`, we don't know which heads were
    used in `forward`, and trying to save the state is messy.  So just make sure that you always
    have conditional logic in `make_output_human_readable` when you implement a `Head`.
    """

    pass



from typing import Dict, Any, Union, Optional

import torch
import numpy as np
from overrides import overrides

from allennlp.data.fields.field import Field
from allennlp.common.util import JsonDict


class TensorField(Field[torch.Tensor]):
    """
    A class representing a tensor, which could have arbitrary dimensions.
    A batch of these tensors are padded to the max dimension length in the batch
    for each dimension.
    """

    __slots__ = ["tensor", "padding_value"]

    def __init__(
        self,
        tensor: Union[torch.Tensor, np.ndarray],
        padding_value: Any = 0.0,
        dtype: Optional[Union[np.dtype, torch.dtype]] = None,
    ) -> None:
        if dtype is not None:
            if isinstance(tensor, np.ndarray):
                tensor = tensor.astype(dtype)
            elif isinstance(tensor, torch.Tensor):
                tensor = tensor.to(dtype)
            else:
                raise ValueError("Did not recognize the type of `tensor`.")
        if isinstance(tensor, np.ndarray):
            tensor = torch.from_numpy(tensor)

        self.tensor = tensor.cpu()
        self.padding_value = padding_value

    @overrides
    def get_padding_lengths(self) -> Dict[str, int]:
        return {"dimension_" + str(i): shape for i, shape in enumerate(self.tensor.size())}

    @overrides
    def as_tensor(self, padding_lengths: Dict[str, int]) -> torch.Tensor:
        tensor = self.tensor
        while len(tensor.size()) < len(padding_lengths):
            tensor = tensor.unsqueeze(-1)
        pad = [
            padding
            for i, dimension_size in reversed(list(enumerate(tensor.size())))
            for padding in [0, padding_lengths["dimension_" + str(i)] - dimension_size]
        ]
        return torch.nn.functional.pad(tensor, pad, value=self.padding_value)

    @overrides
    def empty_field(self):
        # Pass the padding_value, so that any outer field, e.g., `ListField[TensorField]` uses the
        # same padding_value in the padded ArrayFields
        return TensorField(
            torch.tensor([], dtype=self.tensor.dtype), padding_value=self.padding_value
        )

    def __str__(self) -> str:
        return f"TensorField with shape: {self.tensor.size()} and dtype: {self.tensor.dtype}."

    def __len__(self):
        return 1 if len(self.tensor.size()) <= 0 else self.tensor.size(0)

    def __eq__(self, other) -> bool:
        if isinstance(self, other.__class__):
            return (
                torch.equal(self.tensor, other.tensor) and self.padding_value == other.padding_value
            )
        return NotImplemented

    @property
    def array(self):
        """This is a compatibility method that returns the underlying tensor as a numpy array."""
        return self.tensor.numpy()

    @overrides
    def human_readable_repr(self) -> JsonDict:
        shape = list(self.tensor.shape)
        std = torch.std(self.tensor.float()).item()
        mean = torch.mean(self.tensor.float()).item()
        return {
            "shape": shape,
            "element_std": std,
            "element_mean": mean,
            "type": str(self.tensor.dtype).replace("torch.", ""),
        }



"""
A suite of differentiable methods to compute the bias direction
or concept subspace representing binary protected variables.
"""

import torch
import sklearn
import numpy as np

from allennlp.common.checks import ConfigurationError


class BiasDirection:
    """
    Parent class for bias direction classes.

    # Parameters

    requires_grad : `bool`, optional (default=`False`)
        Option to enable gradient calculation.
    """

    def __init__(self, requires_grad: bool = False):
        self.requires_grad = requires_grad

    def _normalize_bias_direction(self, bias_direction: torch.Tensor):
        return bias_direction / torch.linalg.norm(bias_direction)


class PCABiasDirection(BiasDirection):
    """
    PCA-based bias direction. Computes one-dimensional subspace that is the span
    of a specific concept (e.g. gender) using PCA. This subspace minimizes the sum of
    squared distances from all seed word embeddings.

    !!! Note
        It is uncommon to utilize more than one direction to represent a concept.

    Implementation and terminology based on Rathore, A., Dev, S., Phillips, J.M., Srikumar,
    V., Zheng, Y., Yeh, C.M., Wang, J., Zhang, W., & Wang, B. (2021).
    [VERB: Visualizing and Interpreting Bias Mitigation Techniques for
    Word Representations](https://api.semanticscholar.org/CorpusID:233168618).
    ArXiv, abs/2104.02797.
    """

    def __call__(self, seed_embeddings: torch.Tensor):
        """

        # Parameters

        !!! Note
            In the examples below, we treat gender identity as binary, which does not accurately
            characterize gender in real life.

        seed_embeddings : `torch.Tensor`
            A tensor of size (batch_size, ..., dim) containing seed word embeddings related to
            a concept. For example, if the concept is gender, seed_embeddings could contain embeddings
            for words like "man", "king", "brother", "woman", "queen", "sister", etc.

        # Returns

        bias_direction : `torch.Tensor`
            A unit tensor of size (dim, ) representing the concept subspace.
        """

        # Some sanity checks
        if seed_embeddings.ndim < 2:
            raise ConfigurationError("seed_embeddings1 must have at least two dimensions.")

        with torch.set_grad_enabled(self.requires_grad):
            # pca_lowrank centers the embeddings by default
            # There will be two dimensions when applying PCA to
            # definitionally-gendered words: 1) the gender direction,
            # 2) all other directions, with the gender direction being principal.
            _, _, V = torch.pca_lowrank(seed_embeddings, q=2)
            # get top principal component
            bias_direction = V[:, 0]
            return self._normalize_bias_direction(bias_direction)


class PairedPCABiasDirection(BiasDirection):
    """
    Paired-PCA-based bias direction. Computes one-dimensional subspace that is the span
    of a specific concept (e.g. gender) as the first principle component of the
    difference vectors between seed word embedding pairs.

    !!! Note
        It is uncommon to utilize more than one direction to represent a concept.

    Based on: T. Bolukbasi, K. W. Chang, J. Zou, V. Saligrama, and A. Kalai. [Man is to
    computer programmer as woman is to homemaker? debiasing word embeddings]
    (https://api.semanticscholar.org/CorpusID:1704893).
    In ACM Transactions of Information Systems, 2016.

    Implementation and terminology based on Rathore, A., Dev, S., Phillips, J.M., Srikumar,
    V., Zheng, Y., Yeh, C.M., Wang, J., Zhang, W., & Wang, B. (2021).
    [VERB: Visualizing and Interpreting Bias Mitigation Techniques for
    Word Representations](https://api.semanticscholar.org/CorpusID:233168618).
    ArXiv, abs/2104.02797.
    """

    def __call__(self, seed_embeddings1: torch.Tensor, seed_embeddings2: torch.Tensor):
        """

        # Parameters

        !!! Note
            In the examples below, we treat gender identity as binary, which does not accurately
            characterize gender in real life.

        seed_embeddings1 : `torch.Tensor`
            A tensor of size (batch_size, ..., dim) containing seed word
            embeddings related to a concept group. For example, if the concept is gender,
            seed_embeddings1 could contain embeddings for linguistically masculine words, e.g.
            "man", "king", "brother", etc.

        seed_embeddings2: `torch.Tensor`
            A tensor of the same size as seed_embeddings1 containing seed word
            embeddings related to a different group for the same concept. For example,
            seed_embeddings2 could contain embeddings for linguistically feminine words, e.g.
            "woman", "queen", "sister", etc.

        !!! Note
            For Paired-PCA, the embeddings at the same positions in each of seed_embeddings1 and
            seed_embeddings2 are expected to form seed word pairs. For example, if the concept
            is gender, the embeddings for ("man", "woman"), ("king", "queen"), ("brother", "sister"), etc.
            should be at the same positions in seed_embeddings1 and seed_embeddings2.

        !!! Note
            All tensors are expected to be on the same device.

        # Returns

        bias_direction : `torch.Tensor`
            A unit tensor of size (dim, ) representing the concept subspace.
        """

        # Some sanity checks
        if seed_embeddings1.size() != seed_embeddings2.size():
            raise ConfigurationError("seed_embeddings1 and seed_embeddings2 must be the same size.")
        if seed_embeddings1.ndim < 2:
            raise ConfigurationError(
                "seed_embeddings1 and seed_embeddings2 must have at least two dimensions."
            )

        with torch.set_grad_enabled(self.requires_grad):
            paired_embeddings = seed_embeddings1 - seed_embeddings2
            _, _, V = torch.pca_lowrank(
                paired_embeddings,
                q=min(paired_embeddings.size(0), paired_embeddings.size(1)) - 1,
            )
            bias_direction = V[:, 0]
            return self._normalize_bias_direction(bias_direction)


class TwoMeansBiasDirection(BiasDirection):
    """
    Two-means bias direction. Computes one-dimensional subspace that is the span
    of a specific concept (e.g. gender) as the normalized difference vector of the
    averages of seed word embedding sets.

    !!! Note
        It is uncommon to utilize more than one direction to represent a concept.

    Based on: Dev, S., & Phillips, J.M. (2019). [Attenuating Bias in Word Vectors]
    (https://api.semanticscholar.org/CorpusID:59158788). AISTATS.

    Implementation and terminology based on Rathore, A., Dev, S., Phillips, J.M., Srikumar,
    V., Zheng, Y., Yeh, C.M., Wang, J., Zhang, W., & Wang, B. (2021).
    [VERB: Visualizing and Interpreting Bias Mitigation Techniques for
    Word Representations](https://api.semanticscholar.org/CorpusID:233168618).
    ArXiv, abs/2104.02797.
    """

    def __call__(self, seed_embeddings1: torch.Tensor, seed_embeddings2: torch.Tensor):
        """

        # Parameters

        !!! Note
            In the examples below, we treat gender identity as binary, which does not accurately
            characterize gender in real life.

        seed_embeddings1 : `torch.Tensor`
            A tensor of size (embeddings1_batch_size, ..., dim) containing seed word
            embeddings related to a specific concept group. For example, if the concept is gender,
            seed_embeddings1 could contain embeddings for linguistically masculine words, e.g.
            "man", "king", "brother", etc.
        seed_embeddings2: `torch.Tensor`
            A tensor of size (embeddings2_batch_size, ..., dim) containing seed word
            embeddings related to a different group for the same concept. For example,
            seed_embeddings2 could contain embeddings for linguistically feminine words, , e.g.
            "woman", "queen", "sister", etc.

        !!! Note
            seed_embeddings1 and seed_embeddings2 need NOT be the same size. Furthermore,
            the embeddings at the same positions in each of seed_embeddings1 and seed_embeddings2
            are NOT expected to form seed word pairs.

        !!! Note
            All tensors are expected to be on the same device.

        # Returns

        bias_direction : `torch.Tensor`
            A unit tensor of size (dim, ) representing the concept subspace.
        """
        # Some sanity checks
        if seed_embeddings1.ndim < 2 or seed_embeddings2.ndim < 2:
            raise ConfigurationError(
                "seed_embeddings1 and seed_embeddings2 must have at least two dimensions."
            )
        if seed_embeddings1.size(-1) != seed_embeddings2.size(-1):
            raise ConfigurationError("All seed embeddings must have same dimensionality.")

        with torch.set_grad_enabled(self.requires_grad):
            seed_embeddings1_mean = torch.mean(seed_embeddings1, dim=0)
            seed_embeddings2_mean = torch.mean(seed_embeddings2, dim=0)
            bias_direction = seed_embeddings1_mean - seed_embeddings2_mean
            return self._normalize_bias_direction(bias_direction)


class ClassificationNormalBiasDirection(BiasDirection):
    """
    Classification normal bias direction. Computes one-dimensional subspace that is the span
    of a specific concept (e.g. gender) as the direction perpendicular to the classification
    boundary of a linear support vector machine fit to classify seed word embedding sets.

    !!! Note
        It is uncommon to utilize more than one direction to represent a concept.

    Based on: Ravfogel, S., Elazar, Y., Gonen, H., Twiton, M., & Goldberg, Y. (2020).
    [Null It Out: Guarding Protected Attributes by Iterative Nullspace Projection]
    (https://api.semanticscholar.org/CorpusID:215786522). ArXiv, abs/2004.07667.

    Implementation and terminology based on Rathore, A., Dev, S., Phillips, J.M., Srikumar,
    V., Zheng, Y., Yeh, C.M., Wang, J., Zhang, W., & Wang, B. (2021).
    [VERB: Visualizing and Interpreting Bias Mitigation Techniques for
    Word Representations](https://api.semanticscholar.org/CorpusID:233168618).
    ArXiv, abs/2104.02797.
    """

    def __init__(self):
        super().__init__()

    def __call__(self, seed_embeddings1: torch.Tensor, seed_embeddings2: torch.Tensor):
        """

        # Parameters

        !!! Note
            In the examples below, we treat gender identity as binary, which does not accurately
            characterize gender in real life.

        seed_embeddings1 : `torch.Tensor`
            A tensor of size (embeddings1_batch_size, ..., dim) containing seed word
            embeddings related to a specific concept group. For example, if the concept is gender,
            seed_embeddings1 could contain embeddings for linguistically masculine words, e.g.
            "man", "king", "brother", etc.
        seed_embeddings2: `torch.Tensor`
            A tensor of size (embeddings2_batch_size, ..., dim) containing seed word
            embeddings related to a different group for the same concept. For example,
            seed_embeddings2 could contain embeddings for linguistically feminine words, , e.g.
            "woman", "queen", "sister", etc.

        !!! Note
            seed_embeddings1 and seed_embeddings2 need NOT be the same size. Furthermore,
            the embeddings at the same positions in each of seed_embeddings1 and seed_embeddings2
            are NOT expected to form seed word pairs.

        !!! Note
            All tensors are expected to be on the same device.

        !!! Note
            This bias direction method is NOT differentiable.

        # Returns

        bias_direction : `torch.Tensor`
            A unit tensor of size (dim, ) representing the concept subspace.
        """

        # Some sanity checks
        if seed_embeddings1.ndim < 2 or seed_embeddings2.ndim < 2:
            raise ConfigurationError(
                "seed_embeddings1 and seed_embeddings2 must have at least two dimensions."
            )
        if seed_embeddings1.size(-1) != seed_embeddings2.size(-1):
            raise ConfigurationError("All seed embeddings must have same dimensionality.")

        device = seed_embeddings1.device
        seed_embeddings1 = seed_embeddings1.flatten(end_dim=-2).detach().cpu().numpy()
        seed_embeddings2 = seed_embeddings2.flatten(end_dim=-2).detach().cpu().numpy()

        X = np.vstack([seed_embeddings1, seed_embeddings2])
        Y = np.concatenate([[0] * seed_embeddings1.shape[0], [1] * seed_embeddings2.shape[0]])

        classifier = sklearn.svm.SVC(kernel="linear").fit(X, Y)
        bias_direction = torch.Tensor(classifier.coef_[0]).to(device)

        return self._normalize_bias_direction(bias_direction)



"""
*AllenNLP Tango is an experimental API and parts of it might change or disappear
every time we release a new version.*
"""

from allennlp.tango.step import Step
from allennlp.tango.step import MemoryStepCache
from allennlp.tango.step import DirectoryStepCache
from allennlp.tango.step import tango_dry_run
from allennlp.tango.step import step_graph_from_params

from allennlp.tango.training import TrainingStep
from allennlp.tango.evaluation import EvaluationStep

import warnings

warnings.warn(
    "AllenNLP Tango is an experimental API and parts of it might change or disappear "
    "every time we release a new version."
)



from typing import Optional

from fairscale.nn.checkpoint import checkpoint_wrapper
from overrides import overrides
import torch.nn as nn

from allennlp.nn.checkpoint.checkpoint_wrapper import CheckpointWrapper


@CheckpointWrapper.register("fairscale")
class FairScaleCheckpointWrapper(CheckpointWrapper):
    """
    Provides [FairScale]
    (https://fairscale.readthedocs.io/en/latest/api/nn/checkpoint/checkpoint_activations.html)'s
    activation/gradient checkpointing functionality.

    The parameters and their defaults are the same as they are in FairScale, and
    any of them can be overriden on a per-module basis by passing the corresponding parameter
    to `.wrap_module()`.

    This can also be used in conjunction with the
    :class:`allennlp.nn.parallel.fairscale_fsdp_accelerator.FairScaleFsdpAccelerator`.
    See the [T5 implementation](/api/modules/transformer/t5/) for an example
    of how to use the two together.

    !!! Note
        If using the `FairScaleFsdpAccelerator`, you need to set `maintain_forward_counter` to `True`.
        For convenience, if `maintain_forward_counter` is not set, internally it will be
        set to `True` if training in a distributed setup, or `False` otherwise.
    """

    def __init__(
        self, offload_to_cpu: Optional[bool] = True, maintain_forward_counter: Optional[bool] = None
    ) -> None:
        self._offload_to_cpu = offload_to_cpu
        if maintain_forward_counter is None:
            from allennlp.common.util import is_distributed

            # Better to assume we need this in the distributed case, since we definitely
            # need this when the model is wrapped with FairScale's FSDP.
            self._maintain_forward_counter = is_distributed()
        else:
            self._maintain_forward_counter = maintain_forward_counter

    @overrides
    def wrap_module(
        self,
        module: nn.Module,
        **kwargs,
    ) -> nn.Module:
        if "offload_to_cpu" not in kwargs and self._offload_to_cpu is not None:
            kwargs["offload_to_cpu"] = self._offload_to_cpu
        if "maintain_forward_counter" not in kwargs and self._maintain_forward_counter is not None:
            kwargs["maintain_forward_counter"] = self._maintain_forward_counter
        return checkpoint_wrapper(module, **kwargs)
